from unittest import TestCase

from nicetypes.dicts import _DictWrapper, ImmutableDict, TypedDict, validate_dict


class TestValidateDict(TestCase):

    def test_invalid_dict(self):
        # Given
        data = "not_a_dict"
        # When
        with self.assertRaises(TypeError) as context:
            validate_dict(data)
        # Then
        self.assertEqual(
            str(context.exception),
            "Expected a dict as argument, got a <class 'str'> instead."
        )


class TestDictWrapper(TestCase):

    def test_dictwrapper_init(self):
        # Given
        data = {"foo": 1}
        # When
        dwp = _DictWrapper(data)
        # Then
        self.assertDictEqual(data, dwp._inner_dict)  # pylint: disable=protected-access

    def test_dictwrapper_getitem(self):
        # Given
        dwp = _DictWrapper({"foo": 1})
        # When
        value = dwp["foo"]
        # Then
        self.assertEqual(value, 1)

    def test_dictwrapper_str(self):
        # Given
        dwp = _DictWrapper({"foo": 1})
        # When
        dwp_str = str(dwp)
        # Then
        self.assertEqual(dwp_str, "_DictWrapper({'foo': 1})")

    def test_dictwrapper_len(self):
        # Given
        dwp = _DictWrapper({"foo": 1, "bar": 2})
        # When
        dwp_len = len(dwp)
        # Then
        self.assertEqual(dwp_len, 2)

    def test_dictwrapper_get(self):
        # Given
        dwp = _DictWrapper({"foo": 1})
        # When
        value = dwp.get("foo")
        # Then
        self.assertEqual(value, 1)

    def test_dictwrapper_items(self):
        # Given
        dwp = _DictWrapper({"foo": 1})
        # When
        items = tuple(dwp.items())
        # Then
        self.assertSequenceEqual(items, [("foo", 1)])

    def test_dictwrapper_keys(self):
        # Given
        dwp = _DictWrapper({"foo": 1})
        # When
        keys = tuple(dwp.keys())
        # Then
        self.assertSequenceEqual(keys, ["foo"])

    def test_dictwrapper_values(self):
        # Given
        dwp = _DictWrapper({"foo": 1})
        # When
        values = tuple(dwp.values())
        # Then
        self.assertSequenceEqual(values, [1])


class TestImmutableDict(TestCase):

    def test_immutabledict_hashable(self):
        # Given
        imd = ImmutableDict({"foo": 1})
        # When
        hash_ = hash(imd)
        # Then
        self.assertIsInstance(hash_, int)


class TestTypedDict(TestCase):

    def test_typeddict_valid_key(self):
        # Given
        tpd = TypedDict(key_type=str)
        # When
        tpd["one"] = 1
        # Then
        self.assertEqual(tpd["one"], 1)

    def test_typeddict_invalid_key(self):
        # Given
        tpd = TypedDict(key_type=str)
        # When
        with self.assertRaises(TypeError) as context:
            tpd[1] = 1
        # Then
        self.assertEqual(
            str(context.exception),
            "Expected a <class 'str'> as key, got a <class 'int'> instead."
        )

    def test_typeddict_valid_value(self):
        # Given
        tpd = TypedDict(value_type=float)
        # When
        tpd["one"] = 1.0
        # Then
        self.assertEqual(tpd["one"], 1.0)

    def test_typeddict_invalid_value(self):
        # Given
        tpd = TypedDict(value_type=float)
        # When
        with self.assertRaises(TypeError) as context:
            tpd["one"] = 1
        # Then
        self.assertEqual(
            str(context.exception),
            "Expected a <class 'float'> as value, got a <class 'int'> instead."
        )



from unittest import TestCase

from nicetypes.lists import UniqList


class TestUniqList(TestCase):

    def test_uniqlist_init_without_arg(self):
        # Given
        unq = UniqList()
        # When
        list_ = list(unq)
        # Then
        self.assertListEqual(list_, [])

    def test_uniqlist_init_with_tuple(self):
        # Given
        unq = UniqList((1, 2, 3, 3, 4, 1))
        # When
        list_ = list(unq)
        # Then
        self.assertListEqual(list_, [1, 2, 3, 4])

    def test_uniqlist_init_with_list(self):
        # Given
        unq = UniqList([1, 2, 3, 3, 4, 1])
        # When
        list_ = list(unq)
        # Then
        self.assertListEqual(list_, [1, 2, 3, 4])

    def test_uniqlist_init_with_generator(self):
        # Given
        unq = UniqList((i for i in range(1, 5)))
        # When
        list_ = list(unq)
        # Then
        self.assertListEqual(list_, [1, 2, 3, 4])

    def test_uniqlist_getitem(self):
        # Given
        unq = UniqList([1])
        # When
        first = unq[0]
        # Then
        self.assertEqual(first, 1)

    def test_uniqlist_str(self):
        # Given
        unq = UniqList([1, 2, 3])
        # When
        unq_str = str(unq)
        # Then
        self.assertEqual(unq_str, "UniqList([1, 2, 3])")

    def test_uniqlist_append(self):
        # Given
        unq = UniqList([1, 2, 3, 4])
        # When
        unq.append(1)
        unq.append(5)
        # Then
        self.assertListEqual(list(unq), [1, 2, 3, 4, 5])

    def test_uniqlist_insert_replace_false(self):
        # Given
        unq = UniqList([1, 2, 3, 4])
        # When
        unq.insert(0, 3)
        unq.insert(0, 0)
        # Then
        self.assertListEqual(list(unq), [0, 1, 2, 3, 4])

    def test_uniqlist_insert_replace_true(self):
        # Given
        unq = UniqList([1, 2, 3, 4])
        # When
        unq.insert(0, 3, replace=True)
        unq.insert(0, 0, replace=True)
        # Then
        self.assertListEqual(list(unq), [0, 3, 1, 2, 4])

    def test_uniqlist_extend(self):
        # Given
        unq = UniqList([1, 2, 3, 4])
        # When
        unq.extend([4, 5, 6])
        # Then
        self.assertListEqual(list(unq), [1, 2, 3, 4, 5, 6])

    def test_uniqlist_add_uniqlist(self):
        # Given
        unq1 = UniqList([1, 2, 3, 4])
        unq2 = UniqList([4, 5, 6])
        # When
        concat = unq1 + unq2
        # Then
        self.assertListEqual(list(concat), [1, 2, 3, 4, 5, 6])

    def test_uniqlist_add_tuple(self):
        # Given
        unq = UniqList([1, 2, 3, 4])
        # When
        concat = unq + (4, 5, 5, 5, 6)
        # Then
        self.assertListEqual(list(concat), [1, 2, 3, 4, 5, 6])

    def test_uniqlist_add_list(self):
        # Given
        unq = UniqList([1, 2, 3, 4])
        # When
        concat = unq + [4, 5, 5, 5, 6]
        # Then
        self.assertListEqual(list(concat), [1, 2, 3, 4, 5, 6])

    def test_uniqlist_add_generator(self):
        # Given
        unq = UniqList([1, 2, 3, 4])
        # When
        concat = unq + (i for i in range(4, 7))
        # Then
        self.assertListEqual(list(concat), [1, 2, 3, 4, 5, 6])



from copy import deepcopy


def validate_dict(dictionnary):
    if not isinstance(dictionnary, dict):
        raise TypeError("Expected a dict as argument, got a %s instead." % type(dictionnary))


class _DictWrapper:

    def __init__(self, dictionnary):
        validate_dict(dictionnary)
        self._inner_dict = deepcopy(dictionnary)

    def __getitem__(self, key):
        return self._inner_dict[key]

    def __len__(self):
        return len(self._inner_dict)

    def __str__(self):
        return "%s(%s)" % (self.__class__.__name__, str(self._inner_dict))

    def get(self, key):
        return self._inner_dict.get(key)

    def items(self):
        return self._inner_dict.items()

    def keys(self):
        return self._inner_dict.keys()

    def values(self):
        return self._inner_dict.values()


class ImmutableDict(_DictWrapper):

    def __init__(self, dictionnary):
        super().__init__(dictionnary)
        self._hash = None

    def __hash__(self):
        if self._hash is None:
            self._hash = hash(tuple(sorted(self._inner_dict.items())))
        return self._hash


FrozenDict = ImmutableDict


class TypedDict(_DictWrapper):

    def __init__(self, key_type=None, value_type=None, copy_from=None):
        dictionnary = copy_from if copy_from else {}
        super().__init__(dictionnary)
        self._key_type = key_type
        self._value_type = value_type

    def __setitem__(self, key, value):
        self.validate_key(key)
        self.validate_value(value)
        self._inner_dict[key] = value

    def validate_key(self, key):
        if self._key_type and not isinstance(key, self._key_type):
            raise TypeError("Expected a %s as key, got a %s instead." %
                            (self._key_type, type(key)))

    def validate_value(self, value):
        if self._value_type and not isinstance(value, self._value_type):
            raise TypeError("Expected a %s as value, got a %s instead." %
                            (self._value_type, type(value)))
